# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from collections import Counter
from math import acos, factorial, pi, sqrt
from operator import add, floordiv, methodcaller, sub, truediv
from typing import Any, Iterator, List, Optional, Set, Tuple, Type, Union, Collection

from pupy.decorations import cash_it
from pupy.foreign import iter_product



def partitions_gen(numero: int, min_p: int = 1, max_p: Optional[int] = None):
    """Partitions generator

    Adapted from: code.activestate.com/recipes/218332-generator-for-integer-partitions/min_p

    :param numero: number for which to yield partiton tuples
    :type numero: int
    :param min_p: smallest part size (Default value = 1)
    :type min_p: int
    :param max_p: largest part size (Default value = None)
    :type max_p: int

    .. docstring::python

        >>> list(partitions_gen(4))
        [(4,), (1, 3), (1, 1, 2), (1, 1, 1, 1), (2, 2)]
        >>> list(partitions_gen(4, min_p=1, max_p=2))
        [(1, 1, 2), (1, 1, 1, 1), (2, 2)]

    """
    if max_p is None or max_p >= numero:
        yield (numero,)

    for i in range(min_p, numero // 2 + 1):
        for p in partitions_gen(numero - i, i, max_p):
            yield (i,) + p


@cash_it
def rfactorial(n: int) -> int:
    """Recursive factorial function

    :param n:

    .. docstring::python

        >>> from math import factorial
        >>> rfactorial(1) == factorial(1)
        True
        >>> rfactorial(2) == factorial(2)
        True
        >>> rfactorial(3) == factorial(3)
        True
        >>> rfactorial(4) == factorial(4)
        True
        >>> rfactorial(5) == factorial(5)
        True
        >>> rfactorial(6) == factorial(6)
        True
        >>> rfactorial(7) == factorial(7)
        True
        >>> rfactorial(8) == factorial(8)
        True
        >>> rfactorial(9) == factorial(9)
        True


    """
    return 1 if n == 1 else rfactorial(n - 1) * n


def radians_2_degrees(rads: float) -> float:
    """Converts radians to degrees

    :param rads:

    .. doctest:: python

        >>> from math import pi
        >>> radians_2_degrees(2*pi) == 360.0
        True
        >>> radians_2_degrees(2*pi)
        360.0


    """
    return 180 * rads / pi


def degrees_2_radians(degs: float) -> float:
    """Converts degrees to radians

    :param degs:

    .. doctest:: python

        >>> from math import pi
        >>> degrees_2_radians(360.0) == 2*pi
        True

    """
    return degs * pi / 180


def power_mod(number: int, exponent: int, mod: int) -> int:
    """

    :param number:
    :param exponent:
    :param mod:

    .. doctest:: python

        >>> power_mod(2, 4, 3)
        2
        >>> power_mod(12, 3, 4)
        144
        >>> power_mod(123, 2, 3)
        123
        >>> power_mod(120, 6, 10)
        14400
        >>> power_mod(120, 6, 1)
        14400

    """
    if exponent > 0:
        if exponent % 2 == 0:
            return power_mod(number, floordiv(exponent, 2), mod)
        return power_mod(number, floordiv(exponent, 2), mod) * number
    else:
        return 1


def divisors_gen(n: int) -> Iterator[int]:
    """Divisors generator

    :param n: number w/ divisors to be generated
    :type n: int

    .. doctest:: python

        >>> list(divisors_gen(1))
        [1]
        >>> list(divisors_gen(4))
        [1, 2, 4]
        >>> list(divisors_gen(16))
        [1, 2, 4, 8, 16]


    """
    large_divisors = []
    for i in range(1, int(sqrt(n) + 1)):
        if n % i == 0:
            yield i
            if i * i != n:
                large_divisors.append(n // i)
    for divisor in reversed(large_divisors):
        yield divisor


def gcd_it(a: int, b: int) -> int:
    """iterative gcd

    :param a:
    :param b:


        >>> from pupy.maths import gcd_it
        >>> from pupy.maths import gcd_r
        >>> gcd_it(1, 4) == gcd_r(1, 4)
        True
        >>> gcd_it(2, 6) == gcd_r(2, 6)
        True
        >>> gcd_it(3, 14) == gcd_r(3, 14)
        True
        >>> gcd_it(4, 300) == gcd_r(4, 300)
        True

    """
    while a:
        a, b = b % a, a
    return b


@cash_it
def gcd_r(a: int, b: int) -> int:
    """recursive greatest common divisor

    :param a:
    :param b:

    .. doctest:: python

        >>> from pupy.maths import gcd_it
        >>> from pupy.maths import gcd_r
        >>> gcd_it(1, 4) == gcd_r(1, 4)
        True
        >>> gcd_it(2, 6) == gcd_r(2, 6)
        True
        >>> gcd_it(3, 14) == gcd_r(3, 14)
        True
        >>> gcd_it(4, 300) == gcd_r(4, 300)
        True

    """
    if b > a:
        return gcd_r(b, a)
    r = a % b
    if r == 0:
        return b
    return gcd_r(r, b)


def reverse_num(n: int) -> int:
    """Reverses a number

    :param n: number to be reversed
    :type n: int
    :returns: reversed of a number
    :rtype: int

    .. doctest:: python

        >>> reverse_num(1)
        1
        >>> reverse_num(12345)
        54321
        >>> reverse_num(54321)
        12345
        >>> reverse_num(54321000)
        12345

    """

    reversed = 0
    while n > 0:
        reversed *= 10
        reversed += n % 10
        n //= 10
    return reversed


@cash_it
def fib_r(n: int) -> int:
    """Recursively the nth fibonacci number

    :param n: nth fibonacci sequence number
    :type n: int
    :returns: the nth fibonacci number
    :rtype: int

    .. doctest:: python

        >>> fib_r(1)
        1
        >>> fib_r(2)
        2
        >>> fib_r(6)
        13

    """
    return n if n < 3 else fib_r(n - 1) + fib_r(n - 2)


def expo(d: int, n: int) -> int:
    """greatest exponent for a divisor of n

    :param d: divisor
    :type d: int
    :param n: number be divided
    :type n: int
    :returns: number of times a divisor divides n
    :rtype: int

    .. doctest:: python

        >>> expo(100, 2)
        2
        >>> expo(12, 5)
        0
        >>> expo(12, 2)
        2
        >>> expo(160, 4)
        2
        >>> expo(1000, 4)
        1

    """
    if n < d:  # flip
        d, n = n, d
    c = n
    divs = 0
    while c % d == 0:
        c //= d
        divs += 1
    return divs


def pytriple_gen(max_c: int) -> Iterator[Tuple[int, int, int]]:
    """primative pythagorean triples generator

    special thanks to 3Blue1Brown's video on pythagorean triples
    https://www.youtube.com/watch?v=QJYmyhnaaek&t=300s

    :param max_c: max value of c to yeild triples up to
    :type max_c: int

    """
    for real_pts in range(2, int(sqrt(max_c)) + 1, 1):
        for imag_pts in range(real_pts % 2 + 1, real_pts, 2):
            comp = complex(real_pts, imag_pts)
            sqrd = comp * comp
            real = int(sqrd.real)
            imag = int(sqrd.imag)
            if abs(real - imag) % 2 == 1 and gcd_it(imag, real) == 1:
                sea = int((comp * comp.conjugate()).real)
                if sea > max_c:
                    break
                else:
                    yield (imag, real, sea) if real > imag else (real, imag, sea)


def n_permutations_with_replacements(
    it: Union[
        Tuple[int, int, int, int],
        Tuple[int, int, int, int, int, int, int],
        Tuple[int, int, int, int, int],
        Tuple[int, int, int, int, int, int],
    ]
) -> int:
    """

    .. doctest:: python

        >>> n_permutations_with_replacements((1, 2, 3, 4))
        24
        >>> n_permutations_with_replacements((1, 2, 3, 4, 3))
        60
        >>> n_permutations_with_replacements((1, 2, 3, 4, 3, 3))
        120
        >>> n_permutations_with_replacements((1, 2, 3, 4, 3, 4, 4))
        420


    """
    c = Counter(n for n in it)
    a = list(factorial(nc) for nc in c.values())
    return factorial(len(it)) // iter_product(a)


def disjoint(
    a: Union[List[int], List[Union[str, int]]],
    b: Union[List[int], List[Union[str, int]]],
) -> bool:
    """

    :param a:
    :param b:

    .. doctest:: python

        >>> a = [1, 2, 3, 4]
        >>> b = [2, 3, 4, 5]
        >>> disjoint(a, b)
        False
        >>> a = [1, 2, 3, 4]
        >>> b = [5, 6, 7, 8]
        >>> disjoint(a, b)
        True
        >>> a = ['joe', 'frank', 3, 4]
        >>> b = [5, 6, 7, 'frank']
        >>> disjoint(a, b)
        False


    """
    return not any(ae in b for ae in a)


def set_cmp(
    a: Union[Collection[Any], Iterator[Any]],
    b: Union[Collection[Any], Iterator[Any]]
) -> Tuple[Set[Any], Set[Any], Set[Any]]:
    """Compare the elements of two iterables (a and b)

    :param a: first iterable to compare elements of
    :param b: second iterable to compare elements of
    :return: tuple of sets of the form (common elements, in a only, in b only)

    .. doctest::

        >>> a = [n for n in range(6)]
        >>> a
        [0, 1, 2, 3, 4, 5]
        >>> b = list(range(3, 9))
        >>> b
        [3, 4, 5, 6, 7, 8]
        >>> set_cmp(a, b)
        ({3, 4, 5}, {0, 1, 2}, {8, 6, 7})

    """
    if not isinstance(a, set) or not isinstance(b, set):
        return set_cmp(set(a), set(b))
    return a & b, a - b, b - a


def n_choose_r(n, r):
    """

    :param n:
    :param r:

    """
    return factorial(n) // factorial(r) // factorial(n - r)


class Trigon(object):
    """Trigon object composed of three points connected by lines."""

    def __init__(
        self,
        pt1: Union[Tuple[int, int]],
        pt2: Union[Tuple[int, int]],
        pt3: Union[Tuple[int, int]],
    ) -> None:
        self.pt1 = Vuple(pt1)
        self.pt2 = Vuple(pt2)
        self.pt3 = Vuple(pt3)

    @classmethod
    def from_points(cls, pts: List[Tuple[int, int]]):
        """

        :param pts: return:

        """
        if len(pts) == 3:
            return Trigon(*pts)
        if len(pts) == 6:
            it = iter(pts)
            return Trigon(*zip(it, it))

    def __str__(self):
        return "<< {}, {}, {} >>".format(self.pt1, self.pt2, self.pt3)

    def __contains__(self, point: Union[Tuple[int, int]]) -> bool:
        if not isinstance(point, Vuple):
            point = Vuple(point)
        return self.area() == sum(
            map(methodcaller("area"), self.inner_triangles(point))
        )

    def inner_triangles(self, point):
        """Triangle funk that returns the three triangles w/ a point

        The point (p) is connected to each point of a triangle. with points,
        a, b, and c. The three triangles are t1=(a, b, p), t2=(a, c, p), and
        t3 = (b, c, p).

        :param point: point to connect to Triangle Vertices
        :type point: tuple or Vuple
        :returns: t1, t2, t3-> Three triangles

        """
        t1 = Trigon(point, self.pt2, self.pt3)
        t2 = Trigon(self.pt1, point, self.pt3)
        t3 = Trigon(self.pt1, self.pt2, point)
        return t1, t2, t3

    def is_perimeter_point(self, point: Tuple[int, int]) -> bool:
        """

        :param point:

        """
        if not isinstance(point, Vuple):
            point = Vuple(point)
        return any(
            tri_area == 0
            for tri_area in map(methodcaller("area"), self.inner_triangles(point))
        )

    def points(self):
        """ """
        return self.pt1, self.pt2, self.pt3

    def contains_origin(self) -> bool:
        """True if the origin (0,0) lies within the Triangle"""
        return (0, 0) in self

    def area(self) -> float:
        """ """
        return abs(truediv(Vuple.cross(self.pt1 - self.pt2, self.pt3 - self.pt2), 2))

    @staticmethod
    def area_from_points(pt1, pt2, pt3):
        """

        :param pt1:
        :param pt2:
        :param pt3:

        """
        return abs(truediv(Vuple.cross(pt1 - pt2, pt3 - pt2), 2))


class Vuple(tuple):
    """VUPLE == Vector+Tuple"""

    def __new__(cls: Type[Any], *args) -> Any:
        """

        :param args:
        :return:
        """
        return super(Vuple, cls).__new__(cls, tuple(*args))

    def __gt__(self, other: Any) -> bool:
        return Vuple.mag_sqrd(self) > Vuple.mag_sqrd(other)

    def __eq__(self, other: Union[Tuple[float, float], Tuple[int, int], Any]) -> bool:
        return all(a == b for a, b in zip(self, other))

    def __add__(self, k: Union[int, Any]) -> Any:
        """

        .. docstring::python

            >>> va = Vuple((8, 13))
            >>> vb = Vuple((26, 7))
            >>> va + vb
            (34, 20)

        """
        if isinstance(k, int) or isinstance(k, float):
            return Vuple((k + el for el in self))
        elif isinstance(k, Vuple) and len(self) == len(k):
            return Vuple(map(add, self, k))
        raise ValueError("huh idk")

    def __iadd__(self, k):
        return self.__add__(k)

    def __sub__(self, k: Any) -> Any:
        return Vuple(map(sub, self, k))

    def __isub__(self, k):
        return self.__sub__(k)

    def __mul__(self, k: Union[int, Any]) -> Union[int, Any]:
        """Multiply by a scalar for each element or cross product if also iterable of same length

        :param k: scalar ot other iterable to do the cross producting with
        :return: A Vuple or a sum as a cost product

        .. docstring::python

            >>> a = Vuple((1, 2, 3))
            >>> a * a
            14
            >>> a = Vuple((1, 1, 1))
            >>> a * 4
            (4, 4, 4)

        """
        if isinstance(k, int) or isinstance(k, float):
            return self._mul_scalar(k)
        elif isinstance(k, Vuple):
            if len(k) != len(self):
                raise ValueError("Sizes do not match!")
            return Vuple.dot(self, k)

    def __imul__(self, k: int) -> Any:
        return self.__mul__(k)

    def _mul_scalar(self, k: int) -> Any:
        """

        :param k:

        """
        return Vuple((k * el for el in self))

    def __truediv__(self, k: Union[int, float]) -> Any:
        if type(k) is int or type(k) is float:
            return self._truediv_scalar(k)

    def _truediv_scalar(self, k: Union[int, float]) -> Any:
        """

        :param k:

        """
        return Vuple((el / k for el in self))

    def __itruediv__(self, k: int) -> Any:
        return self.__truediv__(k)

    def __floordiv__(self, k):
        if type(k) is int or type(k) is float:
            return self._floordiv_scalar_int(k)

    def __ifloordiv__(self, k):
        return self.__floordiv__(k)

    def _floordiv_scalar_int(self, k):
        """

        :param k:

        """
        return Vuple((el // k for el in self))

    def normalize(self) -> Any:
        """Normalizes the Vuple ST self.magnitude == 1

        :return: Unit Vuple


        """
        return Vuple.unit_vuple(self)

    @staticmethod
    def unit_vuple(voop: Any) -> Any:
        """

        :param voop:

        """
        return Vuple(voop) / Vuple.mag(voop)

    def get_mag_sqrd(self):
        """ """
        return Vuple.mag_sqrd(self)

    @staticmethod
    def mag_sqrd(voop: Union[Tuple[int, int], Any]) -> int:
        """

        :param voop:

        """
        return sum(el * el for el in voop)

    def get_mag(self) -> float:
        """ """
        return Vuple.mag(self)

    @staticmethod
    def mag(voop: Union[Tuple[int, int], Any]) -> float:
        """

        :param voop:

        .. docstring::python

            >>> v = Vuple((3, 4, 5))

        """
        return sqrt(Vuple.mag_sqrd(voop))

    @staticmethod
    def dot(a: Any, b: Any) -> Union[int, float]:
        """

        :param a:
        :param b:

        """
        return sum(va * vb for va, vb in zip(a, b))

    @staticmethod
    def cross(v1: Any, v2: Any) -> int:
        """Cross product of two 2d vectors

        :param v1: first vector
        :param v2: second vector
        :returns: cross product of v1 and v2

        """
        if len(v1) == 2 and len(v2) == 2:
            return (v1[0] * v2[1]) - (v1[1] * v2[0])
        else:
            raise ValueError("cross product gt 2d not implemented")

    @staticmethod
    def angle(v1: Any, v2: Any, radians: bool = False) -> float:
        """

        :param v1:
        :param v2:
        :param radians:  (Default value = False)

        """
        # return acos(Vuple.dproduct(v1, v2)/(Vuple.mag(v1)*Vuple.mag(v2)))
        q = 1 if radians else 180 / pi
        return q * acos(Vuple.dot(Vuple.unit_vuple(v1), Vuple.unit_vuple(v2)))

    def is_disjoint(self, them):
        """

        :param them:

        """
        return disjoint(self, them)

    def product(self) -> int:
        """Multiplies all elements in the Vuple

        :return:

        .. doctest:: pythonm

            >>> v = Vuple((1, 2, 3, 4))
            >>> v.product()
            24
            >>> v = Vuple((100, 1, -1, 2))
            >>> v.product()
            -200
            >>> v = Vuple((100, -1, -1, 2))
            >>> v.product()
            200


        """
        return iter_product(self)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any, Dict, Iterable, List, Union

Flint = Union[int, float]  # float or int
Paths = Iterable[str]  # iterable of path-strings
JASM = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]  # JSON obj



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any, Iterator, List, Optional, Set, Tuple, Type, Union

def partitions_gen(
    numero: int, min_p: int = ..., max_p: Optional[int] = ...
) -> Any: ...
def rfactorial(n: int) -> int: ...
def radians_2_degrees(rads: float) -> float: ...
def degrees_2_radians(degs: float) -> float: ...
def power_mod(number: int, exponent: int, mod: int) -> int: ...
def divisors_gen(n: int) -> Iterator[int]: ...
def gcd_it(a: int, b: int) -> int: ...
def gcd_r(a: int, b: int) -> int: ...
def reverse_num(n: int) -> int: ...
def fib_r(n: int) -> int: ...
def expo(d: int, n: int) -> int: ...
def pytriple_gen(max_c: int) -> Iterator[Tuple[int, int, int]]: ...
def n_permutations_with_replacements(
    it: Union[
        Tuple[int, int, int, int],
        Tuple[int, int, int, int, int, int, int],
        Tuple[int, int, int, int, int],
        Tuple[int, int, int, int, int, int],
    ]
) -> int: ...
def disjoint(
    a: Union[List[int], List[Union[str, int]]],
    b: Union[List[int], List[Union[str, int]]],
) -> bool: ...
def set_cmp(
    a: Union[Set[int], List[int]], b: Union[Set[int], List[int]]
) -> Tuple[Set[int], Set[int], Set[int]]: ...
def n_choose_r(n: Any, r: Any): ...

class Trigon:
    pt1: Any = ...
    pt2: Any = ...
    pt3: Any = ...
    def __init__(
        self,
        pt1: Union[Tuple[int, int]],
        pt2: Union[Tuple[int, int]],
        pt3: Union[Tuple[int, int]],
    ) -> None: ...
    @classmethod
    def from_points(cls: Any, pts: List[Tuple[int, int]]) -> Any: ...
    def __contains__(self, point: Union[Tuple[int, int]]) -> bool: ...
    def inner_triangles(self, point: Any): ...
    def is_perimeter_point(self, point: Tuple[int, int]) -> bool: ...
    def points(self): ...
    def contains_origin(self) -> bool: ...
    def area(self) -> float: ...
    @staticmethod
    def area_from_points(pt1: Any, pt2: Any, pt3: Any): ...

class Vuple(tuple):
    def __new__(cls: Type[Any], *args: Any) -> Any: ...
    def __gt__(self, other: Any) -> bool: ...
    def __eq__(
        self, other: Union[Tuple[float, float], Tuple[int, int], Any]
    ) -> bool: ...
    def __add__(self, k: Union[int, Any]) -> Any: ...
    def __iadd__(self, k: Any): ...
    def __sub__(self, k: Any) -> Any: ...
    def __isub__(self, k: Any): ...
    def __mul__(self, k: Union[int, Any]) -> Union[int, Any]: ...
    def __imul__(self, k: int) -> Any: ...
    def __truediv__(self, k: Union[int, float]) -> Any: ...
    def __itruediv__(self, k: int) -> Any: ...
    def __floordiv__(self, k: Any): ...
    def __ifloordiv__(self, k: Any): ...
    def normalize(self) -> Any: ...
    @staticmethod
    def unit_vuple(voop: Any) -> Any: ...
    def get_mag_sqrd(self): ...
    @staticmethod
    def mag_sqrd(voop: Union[Tuple[int, int], Any]) -> int: ...
    def get_mag(self) -> float: ...
    @staticmethod
    def mag(voop: Union[Tuple[int, int], Any]) -> float: ...
    @staticmethod
    def dot(a: Any, b: Any) -> Union[int, float]: ...
    @staticmethod
    def cross(v1: Any, v2: Any) -> int: ...
    @staticmethod
    def angle(v1: Any, v2: Any, radians: bool = ...) -> float: ...
    def is_disjoint(self, them: Any): ...
    def product(self) -> int: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any, Iterator, List, Optional

from pupy._typing import Flint

def nbytes(num: Flint) -> str: ...
def filesize(filepath: str) -> str: ...
def nseconds(t1: float, t2: Optional[float] = ...) -> str: ...
def term_table(
    strings: List[str], row_wise: bool = ..., filler: str = ...
) -> Iterator[Any]: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from pupy.foreign import (
    chunks,
    digits_list,
    dirs_gen,
    files_gen,
    int_from_digits,
    is_permutation,
    iter_product,
    rotate,
    rotations_gen,
)
from pupy.savings_n_loads import (
    ljson,
    load_jasm,
    lpak,
    lstr,
    lstring,
    safepath,
    save_jasm,
    savings,
    sjson,
    spak,
    sstr,
    sstring,
)
from pupy.sh import cd, cp, ls, mv, parent_dirpath, pwd, rm
from pupy.utils import timestamp



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
"""
========
Pupy CLI
========
"""
from argparse import ZERO_OR_MORE, ArgumentParser
from codecs import decode
from os import makedirs
from os import path as _path

from pupy._template import _IO_PKG_CLI, _IO_SCRIPT, _PKG_INIT, _PKG_MAIN, _PKG_UTILS
from pupy._version import __version__
from pupy.savings_n_loads import sstring
from pupy.sh import pwd


def unescaped_str(arg_str):
    """

    :param arg_str:
    :return:
    """
    return decode(str(arg_str), "unicode_escape")


def new_package(relative_path):
    try:
        makedirs(relative_path, exist_ok=False)
        relative_path
        sstring(
            _path.join(relative_path, "__init__.py"),
            _PKG_INIT.strip("\n").strip("\r\n") + "\n",
        )
        sstring(
            _path.join(relative_path, "__main__.py"),
            _PKG_MAIN.strip("\n").strip("\r\n") + "\n",
        )
        sstring(
            _path.join(relative_path, "cli.py"),
            _IO_PKG_CLI.strip("\n").strip("\r\n") + "\n",
        )
        sstring(
            _path.join(relative_path, "utils.py"),
            _PKG_UTILS.strip("\n").strip("\r\n") + "\n",
        )
        print("Created package: {}".format(relative_path))
    except:
        raise ValueError("{} already exists".format(relative_path))
        return


def new_cmd(args):
    if args.package:
        for relpath in args.path:
            new_package(relpath)
        return

    for fname in args.path:
        if not fname.endswith(".py"):
            fname = fname + ".py"
        new_filepath = _path.join(pwd(), fname)
        sstring(new_filepath, _IO_SCRIPT.strip("\n").strip("\r\n") + "\n")
        print("Created script: {}".format(new_filepath))


PARSER = ArgumentParser(description="Command description.")
SUBPARSERS = PARSER.add_subparsers(help="commands")
PARSER.add_argument("-V", "--version", action="store_true", help="Print pupy version.")

NEW_SUBPARSER = SUBPARSERS.add_parser("new")
NEW_SUBPARSER.add_argument(
    "path", type=unescaped_str, nargs=ZERO_OR_MORE, help="relative path"
)
NEW_SUBPARSER.add_argument(
    "-p", "--package", action="store_true", default=False, help="new package"
)
NEW_SUBPARSER.set_defaults(func=new_cmd)


def main(ARGS=None):
    """

    :param ARGS:
    """

    ARGS = PARSER.parse_args(args=ARGS)
    if ARGS.version:
        print("pupy version: {}".format(__version__))
    if ARGS.func:
        ARGS.func(ARGS)



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from math import ceil
from os import path, stat
from shutil import get_terminal_size
from typing import Any, Iterator, List, Optional

from pupy._typing import Flint


def nbytes(num: Flint) -> str:
    """
    this function will convert bytes to MB.... GB... etc

    .. doctest:: python

        >>> nbytes(100)
        '100.0 bytes'
        >>> nbytes(1000)
        '1000.0 bytes'
        >>> nbytes(10000)
        '9.8 KB'
        >>> nbytes(100000)
        '97.7 KB'
        >>> nbytes(1000000)
        '976.6 KB'
        >>> nbytes(10000000)
        '9.5 MB'
        >>> nbytes(100000000)
        '95.4 MB'
        >>> nbytes(1000000000)
        '953.7 MB'
        >>> nbytes(10000000000)
        '9.3 GB'
        >>> nbytes(100000000000)
        '93.1 GB'
        >>> nbytes(1000000000000)
        '931.3 GB'
        >>> nbytes(10000000000000)
        '9.1 TB'
        >>> nbytes(100000000000000)
        '90.9 TB'

    """
    for x in ["bytes", "KB", "MB", "GB", "TB"]:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0


def filesize(filepath: str) -> str:
    """
    this function will return the file size
    """
    if path.isfile(filepath):
        file_info = stat(filepath)
        return nbytes(file_info.st_size)


def nseconds(t1: float, t2: Optional[float] = None) -> str:
    """Formats time string

    Formats t1 if t2 is None as a string; Calculates the time and formats
    the time t2-t1 if t2 is not None.

    :param t1: time 1/initial in seconds
    :type t1: double
    :param t2: time 2 (Default value = None)
    :type t2: None or double
    :returns: formated string of the t1 - t2 or t1
    :rtype: str

    """
    if t2 is not None:
        return nseconds((t2 - t1))
    elif t1 == 0.0:
        return "0 sec"
    elif t1 >= 1:
        return "%.3f sec" % t1
    elif 1 > t1 >= 0.001:
        return "%.3f ms" % ((10 ** 3) * t1)
    elif 0.001 > t1 >= 0.000001:
        return "%.3f μs" % ((10 ** 6) * t1)
    elif 0.000001 > t1 >= 0.000000001:
        return "%.3f ns" % ((10 ** 9) * t1)
    else:
        return nseconds((t2 - t1))


def term_table(
    strings: List[str], row_wise: bool = False, filler: str = "~"
) -> Iterator[Any]:
    """

    :param strings:
    :param row_wise:
    :param filler:
    :return:
    """
    max_str_len = max(len(str) for str in strings) + 5
    terminal_cols = get_terminal_size((80, 20)).columns
    n_cols = terminal_cols // max_str_len
    n_rows = int(ceil(len(strings) / n_cols))
    spaces = " " * ((terminal_cols - (max_str_len * n_cols)) // n_cols)
    size_string = "{:<" + str(max_str_len) + "}" + spaces
    fmtstring = size_string * (n_cols - 1) + "{:<}"
    strings.extend(filler for _ in range(n_rows * n_cols - len(strings)))
    if row_wise:
        line_iter = zip(*(strings[i::n_cols] for i in range(n_cols)))
    else:
        line_iter = (strings[i::n_rows] for i in range(n_rows))
    return (fmtstring.format(*row) for row in line_iter)



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from os import chdir, getcwd, rename
from typing import Any, List, Optional, Tuple

class LIN:
    @staticmethod
    def rsync_args(
        src: Any,
        dest: Any,
        delete: bool = ...,
        mkdirs: bool = ...,
        exclude: Any = ...,
        include: Any = ...,
        dry_run: bool = ...,
    ) -> List[str]: ...
    @staticmethod
    def rsync(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude: Any = ...,
        include: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def link_dir(linkpath: Any, targetpath: Any) -> None: ...
    @staticmethod
    def link_dirs(link_target_tuples: Any) -> None: ...
    @staticmethod
    def link_file(linkpath: str, targetpath: str) -> None: ...
    @staticmethod
    def link_files(link_target_tuples: Any) -> None: ...
    @staticmethod
    def unlink_dir(link: Any) -> None: ...
    @staticmethod
    def unlink_dirs(links: Any) -> None: ...
    @staticmethod
    def unlink_file(link: Any) -> None: ...
    @staticmethod
    def unlink_files(links: Any) -> None: ...
    sync: Any = ...

class WIN:
    @staticmethod
    def robocopy_args(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude_files: Any = ...,
        exclude_dirs: Any = ...,
        dry_run: bool = ...,
    ) -> List[str]: ...
    @staticmethod
    def robocopy(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude_files: Any = ...,
        exclude_dirs: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def link_dir(link: Any, target: Any) -> None: ...
    @staticmethod
    def link_dirs(link_target_tuples: Any) -> None: ...
    @staticmethod
    def link_file(link: Any, target: Any) -> None: ...
    @staticmethod
    def link_files(link_target_tuples: Any): ...
    @staticmethod
    def unlink_dir(link: Any) -> None: ...
    @staticmethod
    def unlink_dirs(links: Any) -> None: ...
    @staticmethod
    def unlink_file(link: Any) -> None: ...
    @staticmethod
    def unlink_files(links: Any) -> None: ...
    sync: Any = ...

mv = rename
pwd = getcwd
cd = chdir

def cp(
    src: Any, dst: Any, r: bool = ..., symlinks: bool = ..., ignore: Optional[Any] = ...
) -> None: ...
def ls(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_files(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_dirs(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_files_dirs(
    dirpath: str = ..., abs: bool = ...
) -> Tuple[List[str], List[str]]: ...
def rm(*args: Any, r: bool = ...) -> None: ...
def path2name(path_str: str) -> str: ...
def parent_dirpath(fdpath: str) -> str: ...

link_dir: Any
link_dirs: Any
link_file: Any
link_files: Any
unlink_dir: Any
unlink_dirs: Any
unlink_file: Any
unlink_files: Any
sync: Any



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
"""
===========
Async funks
===========
"""
import asyncio
from functools import partial, wraps

from pupy._typing import JASM

try:
    import aiofiles
except ImportError:
    raise ImportError("'pip install aiofiles'..." " if ya wanna use this module!")

try:
    from ujson import dumps
    from ujson import load
    from ujson import loads
except Exception as e:
    from json import dumps
    from json import load
    from json import loads

try:
    from toml import dumps as toml_dumps
    from toml import load as toml_load
    from toml import loads as toml_loads
except ImportError:
    pass

try:
    from ruamel.yaml import YAML

    _yaml_saver = YAML()
    _yaml_loader = YAML(typ="safe")
except ImportError:
    pass


def asyncify(funk):
    @asyncio.coroutine
    @wraps(funk)
    def afunk(*args, loop=None, executor=None, **kwargs):
        loop = loop if loop else asyncio.get_event_loop()
        pfunc = partial(funk, *args, **kwargs)
        return loop.run_in_executor(executor, pfunc)

    return afunk


async def sabytes(filepath: str, _bytes: bytes) -> None:
    """Read bytes from file path

    :param filepath: filepath as as string to read bites from
    :return: some bytes...
    """
    async with aiofiles.open(filepath, mode="wb") as f:
        await f.write(_bytes)


async def labytes(filepath: str) -> bytes:
    """Read bytes from file path

    :param filepath: filepath as as string to read bites from
    :return: some bytes...
    """
    async with aiofiles.open(filepath, mode="rb") as f:
        _bytes = await f.read()
        return _bytes


async def sastring(filepath: str, string: str) -> None:
    try:
        async with aiofiles.open(filepath, mode="w", encoding="utf-8") as f:
            await f.write(string)
    # except NameError:
    #     raise ImportError("'pip install aiofiles' if ya wanna use this!")
    except UnicodeEncodeError:
        async with aiofiles.open(filepath, mode="w", encoding="latin2") as f:
            await f.write(string)


async def lastring(filepath: str) -> str:
    try:
        async with aiofiles.open(filepath, mode="r", encoding="utf-8") as f:
            _file_string = await f.read()
            return _file_string
    except UnicodeDecodeError:
        async with aiofiles.open(filepath, mode="r", encoding="latin2") as f:
            _file_string = await f.read()
            return _file_string


async def sajson(filepath: str, data: JASM, minify: bool = False) -> None:
    """Save json-serial-ize-able data to a specific filepath.

    :param filepath: destination filepath
    :param data: json cereal-izable dictionary/list/thing
    :param minify: Bool flag -- minify the json file
    :return: None

    """
    if type(data) == dict and any(type(val) == bytes for val in data.values()):
        data = {k: str(v, encoding="utf-8") for k, v in data.items()}
    if minify:
        _json_str = dumps(data, ensure_ascii=False)
    else:
        _json_str = dumps(data, indent=4, sort_keys=True, ensure_ascii=False)
    await sastring(filepath, _json_str)


async def lajson(filepath: str) -> JASM:
    """Load a json file given a filepath and return the file-data

    :param filepath: path to the jasm file you want to load
    :return: Loaded file contents
    """
    _json_str = await lastring(filepath)
    return loads(_json_str)


async def latoml(filepath: str) -> JASM:
    try:
        _toml_str = await lastring(filepath)
        return toml_loads(_toml_str)
    except NameError:
        raise EnvironmentError("'pip install toml' if you wanna use this!")


async def satoml(filepath: str, data: JASM) -> None:
    try:
        filepath = filepath if "." in filepath else "{}.toml".format(filepath)
        _toml_str = toml_dumps(data)
        await sastring(filepath, _toml_str)
    except NameError:
        raise EnvironmentError("'pip install toml' if you wanna use this!")


if __name__ == "__main__":
    pass
    # from doctest import testmod
    # testmod()



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python

from typing import Any, Optional

def unescaped_str(arg_str: Any): ...

PARSER: Any

def main(ARGS: Optional[Any] = ...) -> None: ...



# -*- coding: utf-8 -*-
# ~ Jesse K. Rubin ~ Pretty Useful Python
from contextlib import contextmanager
from datetime import datetime
from os import environ, makedirs, path
from shutil import rmtree
from tempfile import mkdtemp
from time import sleep
from typing import Optional
from weakref import finalize

from pupy.sh import cd, link_dirs, link_files, path2name, pwd, unlink_dirs, unlink_files


def timestamp(ts: Optional[float] = None) -> str:
    """Time stamp string w/ format yyyymmdd-HHMMSS

    :return: timestamp string

    .. doctest:: python

        >>> from datetime import datetime
        >>> stamps = ['20190225-161151', '20190225-081151']
        >>> timestamp(1551111111.111111) in stamps
        True
        >>> datetime.now().strftime("%Y%m%d-%H%M%S") == timestamp()
        True

    """
    if ts is None:
        return datetime.now().strftime("%Y%m%d-%H%M%S")
    elif isinstance(ts, float):
        return datetime.fromtimestamp(ts).strftime("%Y%m%d-%H%M%S")
    elif isinstance(ts, datetime):
        return ts.strftime("%Y%m%d-%H%M%S")


def environ_dict():
    return {k: environ[k] for k in environ}


class LinkedTmpDir(object):
    """ make a temp dir and have links."""

    def __init__(
        self,
        suffix=None,
        prefix=None,
        dir=None,
        mkdirs=None,
        lndirs=None,
        lnfiles=None,
        link_targets=None,
    ):
        self.dirpath = mkdtemp(suffix, prefix, dir)
        self.dirname = path.split(self.dirpath)[-1]
        file_targets = []
        dir_targets = []
        if link_targets is None:
            link_targets = []
        for target in link_targets:
            if path.isfile(target):
                file_targets.append(target)
            elif path.isdir(target):
                dir_targets.append(target)
        _rel_file_links = list(map(path2name, file_targets))
        _rel_dir_links = list(map(path2name, dir_targets))

        try:
            assert len(set(_rel_file_links)) == len(_rel_file_links)
        except AssertionError as e:
            raise ValueError("Duplicate filenames present")
        try:
            assert len(set(_rel_dir_links)) == len(_rel_dir_links)
        except AssertionError as e:
            raise ValueError("Duplicate dirnames present")
        _lnk_path = lambda _lnk: path.join(self.dirpath, _lnk)
        self.file_links = list(map(_lnk_path, _rel_file_links))
        self.dir_links = list(map(_lnk_path, _rel_dir_links))
        link_files(zip(self.file_links, file_targets))
        link_dirs(zip(self.dir_links, dir_targets))
        self._finalizer = finalize(
            self,
            self._cleanup,
            self.dirpath,
            warn_message="Implicitly cleaning up {!r}".format(self),
        )

    @classmethod
    def _cleanup(cls, name, warn_message):
        pass

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.dirpath)

    def __enter__(self):
        return self.dirpath

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        unlink_dirs(self.dir_links)
        unlink_files(self.file_links)
        if self._finalizer.detach():
            pass


# from tempfile import TemporaryDirectory, _get_candidate_names, _sanitize_params, mkdtemp
# from pprint import pprint as pp
# from os import path, access, name, fsencode, W_OK, mkdir

# def _mkdtemp(suffix=None, prefix=None, dir=None):
#     """Better version of mkdtemp
#     """
#
#     prefix, suffix, dir, output_type = _sanitize_params(prefix, suffix, dir)
#
#     names = _get_candidate_names()
#     if output_type is bytes:
#         names = map(fsencode, names)
#     for seq in range(10000):
#         name = next(names)
#         file = path.join(dir, prefix + name + suffix)
#         try:
#             mkdir(file, 0o777)
#             # mkdir(file, 0o700)
#         except FileExistsError:
#             continue  # try again
#         except PermissionError:
#             # This exception is thrown when a directory with the chosen name
#             # already exists on windows.
#             if (name == 'nt' and path.isdir(dir) and
#                 access(dir, W_OK)):
#                 continue
#             else:
#                 raise
#         return file
#
#     raise FileExistsError("No usable temporary directory name found")


@contextmanager
def linked_tmp_dir(
    suffix=None, prefix=None, dir=None, mkdirs=[], lndirs=[], lnfiles=[]
):
    temp_dir = mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    lnfiles = [
        (path.join(temp_dir, _rel_link), target) for _rel_link, target in lnfiles
    ]
    lndirs = [(path.join(temp_dir, _rel_link), target) for _rel_link, target in lndirs]
    # print(mkdirs)
    _dirs2make = [
        path.join(temp_dir, e)
        for e in (
            dirpath if isinstance(dirpath, str) else path.join(*dirpath)
            for dirpath in mkdirs
        )
    ]
    _dirs2make.extend((path.split(link)[0] for link, target in lnfiles))
    _dirs2make.extend((path.split(link)[0] for link, target in lndirs))
    for dirpath_route in _dirs2make:
        # print("mkingdir", dirpath_route)
        makedirs(path.join(temp_dir, dirpath_route), exist_ok=True)

    link_files(lnfiles)
    link_dirs(lndirs)
    # from pupy.foreign import files_gen, dirs_gen
    # from pprint import pprint
    # pprint(list(files_gen(temp_dir)))
    # pprint(list(dirs_gen(temp_dir)))
    # try:
    #     lndirs = (
    #         (path.join(temp_dir, _rel_link), target) for _rel_link, target in lndirs
    #     )
    # except TypeError as e:
    #     pass
    try:
        yield temp_dir
    finally:
        try:
            unlink_files(lnfiles)
        except Exception as e:
            pass
        try:
            unlink_dirs(lndirs)
        except Exception as e:
            pass
        try:
            rmtree(temp_dir)
        except PermissionError:
            # sleep(3)
            # print(pwd())
            # print(temp_dir)
            cd("..")
            # print(pwd())
            rmtree(temp_dir)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pupy']

package_data = \
{'': ['*']}

extras_require = \
{'io': ['toml>=0.10.0,<0.11.0',
        'ujson>=1.35,<2.0',
        'msgpack>=0.6.1,<0.7.0',
        'ruamel.yaml>=0.15.97,<0.16.0',
        'aiofiles>=0.4.0,<0.5.0']}

entry_points = \
{'console_scripts': ['pupy = pupy.cli:main']}

setup_kwargs = {
    'name': 'pupy',
    'version': '2.16.5',
    'description': 'pretty useful python',
    'long_description': "# PUPY ~ Pretty Useful Python\n\nThis package (pupy) is full of Pretty Useful Python.\n\n## These things\n\n[![Build Status](https://travis-ci.org/jessekrubin/pupy.svg?branch=master)](https://travis-ci.org/jessekrubin/pupy) \n[![Wheel](https://img.shields.io/pypi/wheel/pupy.svg)](https://img.shields.io/pypi/wheel/pupy.svg) \n[![Version](https://img.shields.io/pypi/v/pupy.svg)](https://img.shields.io/pypi/v/pupy.svg) \n[![py_versions](https://img.shields.io/pypi/pyversions/pupy.svg)](https://img.shields.io/pypi/pyversions/pupy.svg) \n \n\n## Installin'\n\nInstalling is EZPZ...\n\n    pip install pupy\n\nIf you plan on using Jason you can install `ujson` for a speed boost *(if you wanna go fast; pupy uses the python-stdlib json if you wanna go slow)*...\n\n    pip install ujson \n\nBut wait theres more! Pupy has several optional dependencies for savings and loads which you can install with...\n    \n    pip install ruamel.yaml  # If ya mel in your spare time\n    pip install msgpack  # If you wanna pack messages\n    pip install toml  # toml (aka 'that other markup language'(?))\n    pip install aiofiles  # if you wanna do async stuff\n\n## Testimonials\n\n*'Ahead of the curve.' -Haley who goes to Harvard*\n\n*'huh?' -Tim Apple*\n\n*'If I were stranded on a dessert island and had to pick between having a copy pupy or a thing of water I would pick the water, but if I didn't have to pick it wouldn't hurt to have.' -Ryan of the fro with his fro*\n\n*'I can't say google wouldn't not go down if it weren't for pupy.' -Ben W from the g-suite*\n\n*'A literary masterpiece on par with my second greatest work finite jest.' -Bananas Foster Wallace*\n\n*'Best place for python savings and loads in the tri-content area.' -Genghis Khan*\n",
    'author': 'jessekrubin',
    'author_email': 'jessekrubin@gmail.com',
    'url': 'https://github.com/jessekrubin/pupy',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)
